# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Python utilities

try: # permet d'importer les .so de mkl à partir d'un module python
    import sys,DLFCN
    sys.setdlopenflags(DLFCN.RTLD_NOW|DLFCN.RTLD_GLOBAL)
except:
    pass

# ------------------------------------------------------------------------------

def findbins(modname, dirB=['build','wavesB'], verb=False): # changer wavesB en autre chose !
    """
    try to find where the binaries related to a python module are
    if modname='fwk' => look for 'fwkw.[so|pyd]'
    """

    modwrap = modname+'w' # assumption: SWIG-generated module = module+'w'

    # are we lucky?
    try:
        __import__(modwrap)   #import fwkw
        if verb:
            print("[findbins] module %s already in the PYTHONPATH!" % modwrap)
        return # nothing to do
    except:
        pass
    if verb:
        print("[findbins] module %s is not in the PYTHONPATH!" % modwrap)

    # the module is not in the pythonpath
    # let's look for a folder called "dirB" in parent directories
    # result in "spath"
    import sys, os, os.path
    path = os.path.abspath(sys.modules[modname].__file__)
    found = False
    while True:
        for dir in dirB:
            spath = os.path.join(path,dir)
            if verb:
                print("[findbins] path = ", path)
                print("[findbins] spath = ", spath)
                print("[findbins] os.path.isdir(spath) = ", os.path.isdir(spath))
            if(os.path.isdir(spath)):
                found = True
                if verb:
                    print("[findbins] break with ",dir)
                break
        uppath = os.path.split(path)[0]
        if(path==uppath):
            raise Exception('[findbins] binary dir not found in '+repr(dirB)+' !')
        path=uppath

        if found == True : # to break the while ...
            break
    if verb:
        print("[findbins] spath=", spath)


    # build a list of possible directories (libpath)
    # according to the system and the runtime environment

    pyexe = os.path.basename(sys.executable)
    if verb:
        print("[findbins] pyexe=", pyexe)

    libpath = []
    if pyexe.find('_d.exe')>=0:
        print("[findbins] looking for win/debug libs")
        libpath.append(os.path.join(spath, 'bin/Debug'.replace('/',os.sep)))
        if 'GCC' in sys.version: # windows+gcc = mingw => needs to append lib dir to PATH
            os.environ['PATH'] += ';' + os.path.join(spath, 'bin')
    elif pyexe.find('.exe')>=0:
        print("[findbins] looking for win/release libs")
        libpath.append(os.path.join(spath, 'bin/Release'.replace('/',os.sep)))
        libpath.append(os.path.join(spath, 'bin/RelWithDebInfo'.replace('/',os.sep)))
        if 'GCC' in sys.version: # windows+gcc = mingw => needs to append lib dir to PATH
            os.environ['PATH'] += ';' + os.path.join(spath, 'bin')
    else:
        print("[findbins] looking for linux/mingw libs")
    libpath.append( os.path.join(spath, 'bin') ) # also for win-mingw

    # check the paths one by one
    for p in libpath:
        if os.path.isdir(p):
            if verb:
                print("[findbins] adding %s to sys.path" % p)
            sys.path.append(p)
            if verb:
                print("[findbins] sys.path =", sys.path)
            break
    else:
        raise Exception("[findbins] '%s' not found!" % libpath)

    #exec("import %s" % modwrap)
    #mod = sys.modules[modwrap]
    mod = __import__(modwrap)
    print('[findbins] loading ', mod.__file__)

# attention! si waves and heat possedent des classes de mêmes noms,
# c'est celle du premier module qui sera chargée!
# => ne pas charger waves et heat en même temps tant qu'il n'y a pas de namespaces
#    distingant les 2 ensembles de classes

#-------------------------------------------------------------------------------

def initDLL():
    # For new versions of python, DLLs must be loaded via a specific command
    import platform, os, sys
    if 'Windows' in platform.uname() and sys.version_info.minor >= 8:
        lookfor = ['tbb', 'mkl', 'vtk', 'mumps', 'compiler']
        for k in lookfor:
            print(f'[initDLL] looking for {k}')
            found=False
            for v in os.environ['path'].split(';'):
                if k in v.lower():
                    print(f'[initDLL]\tos.add_dll_directory("{v}")')
                    os.add_dll_directory(v)
                    found=True
                    break
        if not found:
            print('[initDLL]**WARNING: {k} dll path not found!')

# ------------------------------------------------------------------------------

def initMKL(verb=False):
    """Initialize the environment for MKL
       Since we use the SDL for MKL, we need to configure it at runtime.
       This is MUST be performed before any calls to MKL.
       Could be improved by performing the init from the C++ and actually checking that the threading layer is the one we want, instead of using env var.
    """
    import os
    # Define threading interface
    if not 'MKL_INTERFACE_LAYER' in os.environ:
        os.environ['MKL_INTERFACE_LAYER'] = 'LP64' # 32 bits (default), ILP64 for 64 bits
    # Define threading layer
    if not 'MKL_THREADING_LAYER' in os.environ:
        os.environ['MKL_THREADING_LAYER'] = 'TBB' # sequential computation will be performed with 1 TBB thread
    # Force number of OMP threads to 1, TBB threads are controlled from inside the C++ code
    if not 'MKL_NUM_THREADS' in os.environ:
        os.environ['MKL_NUM_THREADS'] = '1'
        os.environ['MKL_DOMAIN_NUM_THREADS'] = 'MKL_DOMAIN_ALL=1'
    # Force number of OMP threads to remain unchanged
    if not 'MKL_DYNAMIC' in os.environ:
        os.environ['MKL_DYNAMIC'] = 'FALSE'
    # Display variables
    if verb:
        for s in ['OMP', 'MKL', 'KMP']:
            print('* %s environment:' % s)
            for key, value in os.environ.items():
                if s+'_' in key:
                    print('    %s = %s' % (key, value))
    print('')

# ------------------------------------------------------------------------------

def setupwdir(testname):
    """
    creates a single working folder for the given test
    """
    try:
        # are we using MPI?
        import mpi4py.MPI as mpi
        comm = mpi.COMM_WORLD
        rank = comm.rank
        siz  = comm.size
        name = mpi.Get_processor_name()
        status = mpi.Status()
    except:
        comm = None
        rank = 0
        siz  = 1
        name = "noname"
    import os, os.path

    # builds the name of the workspace folder
    #print "__file__=",__file__
    dir1 = os.path.abspath(os.path.dirname(__file__)+os.sep+"..")+os.sep
    print("dir1=",dir1)
    print("testname=",testname)
    common = os.path.commonprefix( (testname, dir1) )
    #print "common=", common
    resdir = testname[len(common):].replace(os.sep,"_")
    resdir = os.path.splitext(resdir)[0] # remove ".py"
    if siz>1:
        # append the mpi size if mpi is used
        # (so that the same test can be run concurrently with
        # sevral mpi sizes in ctest)
        resdir += "_mpi%d" % siz
    #print "resdir=", resdir
    wdir = os.path.join('workspace', resdir)

    if rank == 0:
        # rank0 is in charge of creating the folder...
        args = parseargs()
        if not os.path.isdir(wdir):
            print("creating", wdir)
            os.makedirs(wdir)
        elif os.path.isdir(wdir) and args.clean:
            print('cleaning', wdir)
            import shutil
            for f in os.listdir(wdir):
                fpth = os.path.join(wdir, f)
                if os.path.isfile(fpth):
                    os.remove(fpth)
                elif os.path.isdir(fpth):
                    shutil.rmtree(fpth)

    if siz>1: # if multiple process => send sync to slaves.
        comm.Barrier()

    # we can now safely go into the new folder
    os.chdir(wdir)

# ------------------------------------------------------------------------------

def parseargs():
    """
    parses command line arguments
    """
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verb", help="increase output verbosity", action="count", default=0)
    parser.add_argument("--nogui", help="disable any graphical output", action="store_true")
    parser.add_argument("--fpe", help="enable floating point exceptions", action="store_true")
    parser.add_argument("--clean", help="clean workspace", action="store_true")
    parser.add_argument("-k", help="nb of threads", type=int, default=1)
    parser.add_argument('file', help='python files')
    args = parser.parse_args()
    return args

# ------------------------------------------------------------------------------

def pth(filename, caller=''):
    import os, os.path
    retname = filename
    # try to build the name of the file if a caller file or dir is given
    if not os.path.isabs(filename):
        if caller:
            #print "caller=",caller
            if os.path.isfile(caller):
                dirname = os.path.split(caller)[0]
            elif os.path.isdir(caller):
                dirname = caller
            else:
                raise Exception("neither a file nor a dir (%s)" %caller)
            f = os.path.join(dirname, filename)
            if os.path.isfile(f):
                retname = f
    # check if the file exist
    if not os.path.isfile(retname):
        raise Exception("path to a non-existent file! (%s)" % filename)
    #print "retname=",retname
    return retname

# ------------------------------------------------------------------------------

def run():
    global __file__ # required when run.py is called from an external directory
    global __name__
    import os
    import fwk
    from fwk.streams import Tee
    from fwk.coloring import ccolors

    # redirect C++ streams to Python
    redirect = fwk.StdOutErr2Py()

    # parse args
    args = parseargs()
    initMKL(verb=True) # initialize MKL environment (threading interface, layer and number of threads)
    if args.fpe:
        fwk.enableFpe() # enables floating point exceptions at runtime
        print(ccolors.ANSI_YELLOW + 'Floating point exceptions will cause numpy to overflow!' + ccolors.ANSI_RESET)

    # prepare test
    testname = os.path.abspath(args.file)
    if not os.path.isfile(testname):
        raise Exception("file not found: %s" % testname)
    setupwdir(testname)
    __file__ = testname
    __name__ = "__main__"
    print("[run.py] __file__", __file__)
    tee = Tee('stdout.txt') # split streams

    # start test
    import time, socket
    print('-' * 80)
    print(ccolors.ANSI_BLUE + 'PyStarting...' + ccolors.ANSI_RESET)
    print("starting test", testname)
    print("time:", time.strftime("%c"))
    print("hostname:", socket.gethostname())
    script = open(testname, 'r', encoding='utf-8').read()
    exec(compile(script, testname, 'exec'), {'__file__': testname, '__name__':'__main__'}) # compile to get clean backtrace
