/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WFCT0_H
#define WFCT0_H

#include "tbox.h"
#include "wObject.h"
#include <vector>
#include <map>
#include <Eigen/Dense>

namespace tbox
{

class TBOX_API PwLf : public fwk::wSharedObject
{
    std::map<double, double> pts;

public:
    PwLf() : wSharedObject() {}
    void add(double x, double y);
    double eval(double x) const;
#ifndef SWIG
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Scalar functions to be integrated over an element
 *
 * eval() evaluates the function at kth Gauss point
 * @authors Romain Boman
 */
class TBOX_API Fct0 : public fwk::wSharedObject
{
public:
#ifndef SWIG
    Fct0() : wSharedObject()
    {
    }
#endif
    virtual ~Fct0()
    {
    }

#ifndef SWIG
    virtual double eval(Element const &e,
                        std::vector<double> const &u,
                        size_t k) const = 0;
#endif
};

/**
 * @brief Constant function
 * @authors Romain Boman
 */
class TBOX_API Fct0C : public Fct0
{
    double v;

public:
    Fct0C(double _v) : Fct0(), v(_v) {}

#ifndef SWIG
    virtual double eval(Element const &e,
                        std::vector<double> const &u,
                        size_t k) const override
    {
        return v;
    }
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief Function depending on position
 * @authors Romain Boman
 */
class TBOX_API Fct0XYZ : public Fct0
{
public:
    virtual double eval(Eigen::Vector3d const &pos) const = 0;

#ifndef SWIG
    virtual double eval(Element const &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

/**
 * @brief function depending on element
 * @authors Romain Boman
 */
class TBOX_API Fct0U : public Fct0
{
public:
    virtual double eval(double u) const = 0;

#ifndef SWIG
    virtual double eval(Element const &e,
                        std::vector<double> const &u,
                        size_t k) const override;
    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace tbox

#endif // WFCT0_H
