/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef WGMRES_H
#define WGMRES_H

#include "tbox.h"
#include "wLinearSolver.h"
#include <unsupported/Eigen/IterativeSolvers>

namespace tbox
{

/**
 * @brief GMRES interface
 * @authors Adrien Crovato
 */
class TBOX_API Gmres : public LinearSolver
{
    int fill;   // fill-in factor for ILUT
    double tsh; // drop tolerance threshold for ILUT
    int rst;    // number of iterations before restart for GMRES
    double tol; // relative tolerance for GMRES
    int mit;    // maximum number of iterations for GMRES
    Eigen::GMRES<Eigen::SparseMatrix<double>, Eigen::IncompleteLUT<double>> solver;

public:
    Gmres(int _fill = 1, double _tsh = 1e-6, int _rst = 30, double _tol = 1e-8, int _mit = 1000);
    Gmres(const Gmres &gmres) : Gmres(gmres.fill, gmres.tsh, gmres.rst, gmres.tol, gmres.mit) {}
    virtual ~Gmres() {}
    virtual LSolType type() const override { return LSolType::GMRES_ILUT; }

#ifndef SWIG
    virtual void analyze(Eigen::SparseMatrix<double> const &A) override;
    virtual void factorize(Eigen::SparseMatrix<double> const &A) override;
    virtual void solve(Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x) override;
    virtual void compute(Eigen::SparseMatrix<double> const &A, Eigen::Map<Eigen::VectorXd> const &b, Eigen::Map<Eigen::VectorXd> &x) override;

    virtual double getError() override { return solver.error(); }
    virtual int getIterations() override { return static_cast<int>(solver.iterations()); }

    void setTolerance(double t);

    virtual void write(std::ostream &out) const override;
#endif
};

} // namespace tbox

#endif // WGMRES_H
